@@ -4,14 +4,25 @@ import pkgutil
 import re
 import wikipediaapi
 
+from .foodcategory import CategorySynset
 from .foodcategory import FoodCategory
 from collections import Counter
+from nltk.corpus import wordnet
 from PyDictionary import PyDictionary
 from pyfood.utils import Shelf
 from textblob import Word
 
 embedding_path = 'data/ingredient_embedding.npy'
 
+def __get_embedding():
+    """
+    Get the dataset of ingredients as a dictionary.
+
+    :return: a dictionary representing the embedding
+    """
+    embedding_io = io.BytesIO(pkgutil.get_data(__name__, embedding_path))
+    return np.load(embedding_io, allow_pickle=True).item()
+
 def lemmatize_word(word):
     """
     Lemmatize the provided word.
@@ -30,6 +41,7 @@ def is_ingredient_vegan(ingredient):
     :param ingredient: the name of the ingredient.
     :return: a bool indicating whether the ingredient is vegan or not.
     """
+    ingredient = ingredient.strip()
     shelf = Shelf('Milan', month_id=0)
     results = shelf.process_ingredients([ingredient])
     return results['labels']['vegan']
@@ -46,6 +58,55 @@ def is_recipe_vegan(ingredients):
     results=shelf.process_ingredients(ingredients)
     return results['labels']['vegan']
 
+def add_ingredient(ingredient, tag):
+    """
+    Map the provided ingredient and the tag into the embedding dataset. 
+    Tag must be one the following FoodCategory: 
+    vegetable, fruit, meat, legume, diary, egg, staple, condiment, nut, seafood, dessert
+
+    :param ingredient: the name of the ingredient.
+    :param tag: the class of the ingredient. Must be one of the listed above.
+    :return: a bool indicating if the operation has succeded or not. 
+    """
+    embedding = __get_embedding()
+    ingredient = ingredient.strip()
+    tag = tag.strip()
+    if ingredient in embedding:
+        return False
+    
+    embedding[ingredient] = FoodCategory[tag].value
+    return True
+
+def search_ingredient_hypernyms(ingredient):
+    """
+    Predict the class of the provided ingredient based on the Wu & Palmer’s similarity between
+    ingredient, his hypernyms and the 11 FoodCategory. 
+    The FoodCategory is choosen based on the maximum similarity value between the ingredient, 
+    its hypernym and the various categories. If the predicted category is different between ingredient
+    and hypernym the category is choosen based on the avarege of both. 
+
+    :param ingredient: the name of the ingredient.
+    :return: the class of the ingredient.
+    """
+    ingredient = wordnet.synset(f'{ingredient}.n.01')
+    hypernym = ingredient.hypernyms()[0]
+    categories = CategorySynset.categories
+
+    sim = []
+    hypernym_sim = []
+    for cat in categories:
+        sim.append(ingredient.wup_similarity(cat))
+        hypernym_sim.append(hypernym.wup_similarity(cat))
+    
+    best_sim = sim.index(max(sim))
+    best_hyp = hypernym_sim.index(max(hypernym_sim))
+
+    if best_sim == best_hyp: 
+        return FoodCategory(best_sim).name
+    else: 
+        sum = [(x + y)/2 for x, y in zip(sim, hypernym_sim)]
+        return FoodCategory(sum.index(max(sum)))
+
 def search_ingredient_class(ingredient):
     """
     Search on wikipedia and english dictionary the class of the provided ingredient.
@@ -66,26 +127,25 @@ def search_ingredient_class(ingredient):
             categories.append(category.name)
         if re.search(r'\b({0})\b'.format(category.name), ontology):
             categories.append(category.name)
-    if categories:
-        return max(categories,key=categories.count) 
-    else:
-        return categories
+    return max(categories, key=categories.count) if len(categories) else None
 
 def get_ingredient_class(ingredient):
     """
     Predict the class of the provided ingredient based on the embeddings. 
-    If the ingredient cannot be found in the dictionary it will be searched on wikipedia pages.
+    If the ingredient cannot be found in the dictionary it will be searched on wikipedia pages or hypernyms.
 
     :param ingredient: the name of the ingredient.
     :return: the class of the ingredient.
     """
-    embedding = io.BytesIO(pkgutil.get_data(__name__, embedding_path))
-    embedding = np.load(embedding, allow_pickle=True).item()
-    if ingredient in embedding:
-        lemmatized_ing = lemmatize_word(ingredient)
+    embedding = __get_embedding()
+    ingredient = ingredient.strip()
+    lemmatized_ing = lemmatize_word(ingredient)
+    if lemmatized_ing in embedding:
         return FoodCategory(embedding[lemmatized_ing]).name
     else:
-        return search_ingredient_class(ingredient)
+        web_class = search_ingredient_class(ingredient)
+        hyp_class = search_ingredient_hypernyms(lemmatized_ing)
+        return web_class if web_class else hyp_class
 
 def get_recipe_class_percentage(ingredients):
     """
@@ -102,9 +162,11 @@ def get_recipe_class_percentage(ingredients):
 def get_recipe_tags(ingredients):
     """
     Classify a recipe in tags based on its ingredient. 
-    Tag could be: Vegetable, Fruit, Meat, Legume, Diary, Egg. 
+    Tag could be: Vegetable, Fruit, Meat, Legume, Diary, Egg, Staple, Condiment, Nut, Seafood 
 
     :param ingredients: list of ingredients in the recipe.
     :return: set of tags for the recipe. 
     """
-    return list(set([get_ingredient_class(ingredient) for ingredient in ingredients]))               
\ No newline at end of file
+    tags = [get_ingredient_class(ingredient) for ingredient in ingredients]
+    if None in tags: tags.remove(None)
+    return list(set(tags)) if len(tags) else tags    
\ No newline at end of file
